from .....schema_classes import CorpGroupInfoClass
from .....schema_classes import CorpUserEditableInfoClass
from .....schema_classes import CorpUserInfoClass


CorpGroupInfo = CorpGroupInfoClass
CorpUserEditableInfo = CorpUserEditableInfoClass
CorpUserInfo = CorpUserInfoClass
